
// Copyright (C) 2008 Indrek Mandre <indrek@mare.ee>
// http://www.mare.ee/indrek/flashme/

#include <avr/boot.h>
#include <avr/wdt.h>

#include "xavr/xavr.h"

typedef XAVR::PortC::Out4 BLINK;

#define FREQ 8000000

static void USART0Init (uint16_t brate)
{
  UBRR0 = FREQ / 16 / brate - 1;
  // enable receive and transmit
  UCSR0B = _BV(RXEN0) | _BV(TXEN0);
  // asyncronous, no parity, 1 stop bit, 8 bit character size
  UCSR0C = _BV(UCSZ00) | _BV(UCSZ01);
}

static void USARTSend (const char *str)
{
  while (*str)
    {
      while ( ! (UCSR0A & _BV(UDRE0)) );
      UDR0 = *str++;
    }
}

static int16_t USARTGetc (bool block = true)
{
  if ( block )
    {
      while ( !(UCSR0A & (1<<RXC0)) );
      return UDR0;
    }
  else
    {
      if ( !(UCSR0A & (1<<RXC0)) )
          return -1;
      return UDR0;
    }
}

extern "C" bool run_self_flashing () __attribute__ ((section ("flashprog")));

bool run_self_flashing ()
{
  uint16_t len = 0;
  uint16_t page = 0;

  BLINK::Clear();

  // at first read the binary size to be written into the flash (in bytes)
  while ( !(UCSR0A & (1<<RXC0)) );
  len |= UDR0;
  while ( !(UCSR0A & (1<<RXC0)) );
  len |= UDR0 << 8;

  BLINK::Set();

  if ( len >= 3000 ) // don't allow overwriting our boot section or insane values
      return false;

  // disable interrupts
  cli();

  // make sure no pending eeprom writes are there
  eeprom_busy_wait ();

  while (len != 0)
    {
      // at first erase the given page
      boot_page_erase (page);
      boot_spm_busy_wait (); // wait for erase to complete

      // ask for more data
      while ( ! (UCSR0A & _BV(UDRE0)) );
      UDR0 = 'N';
      while ( ! (UCSR0A & _BV(UDRE0)) );
      UDR0 = SPM_PAGESIZE & 0xff;
      while ( ! (UCSR0A & _BV(UDRE0)) );
      UDR0 = (SPM_PAGESIZE & 0xff00) >> 8;

      // now fill the buffer page
      for ( size_t i = 0; i < SPM_PAGESIZE; i += 2 )
        {
          uint16_t w = 0;
          if ( len )
            {
              while ( !(UCSR0A & (1<<RXC0)) );
              w = UDR0;
              len--;
            }

          if ( len )
            {
              while ( !(UCSR0A & (1<<RXC0)) );
              w |= UDR0 << 8;
              len--;
            }

          // write the word into the page buffer
          boot_page_fill (page + i, w);
        }

      boot_page_write (page); // write the page to flash
      boot_spm_busy_wait();   // wait for the write to complete

      page += SPM_PAGESIZE;
    }

  // jump to beginning, virtual reset
#define START0() (((void(*)(void))(char *)0x0000)())
  START0();

  return true;
}

int main ()
{
  BLINK::Init();

  USART0Init (9600);

  const char *str1 = "Hello!\r\n";
  const char *str2 = "Goodbye!\r\n";
  const char *send = str1;

  while(1)
    {
      BLINK::Toggle();

      for ( uint32_t i = 0; i < 250000; i++ )
          __asm__ __volatile__ ("nop");
      USARTSend (send);

      while (1)
        {
          int16_t c = USARTGetc(false);
          if ( c == -1 )
              break;
          if ( c == 'h' || c == 'H' )
              send = str1;
          else if ( c == 'g' || c == 'G' )
              send = str2;
          else if ( c == 'f' || c == 'F' )
            {
              if ( !run_self_flashing() )
                {
                  BLINK::Clear();
                  while(1);
                }
            }
        }
    }
  return 0;
}

